@@ -354,8 +354,13 @@ SMTP_RESP *smtp_chat_resp(SMTP_SESSION *session)
 	 * loss of mail is not acceptable then they can turn off pipelining
 	 * for specific sites, or they can turn off pipelining globally when
 	 * they find that there are just too many broken sites.
+	 * 
+	 * Fix 20190621: don't cache an SMTP session after an SMTP protocol
+	 * error. The protocol may be in a bad state. Disable caching here so
+	 * that the protocol engine will send QUIT.
 	 */
 	session->error_mask |= MAIL_ERROR_PROTOCOL;
+	DONT_CACHE_THIS_SESSION;
 	if (session->features & SMTP_FEATURE_PIPELINING) {
 	    msg_warn("%s: non-%s response from %s: %.100s",
 		     session->state->request->queue_id,
